# coding: utf-8
"""Extension properties for Building, Story, Room2D.

These objects hold all attributes assigned by extensions like dragonfly-radiance
and dragonfly-energy.  Note that these Property objects are not intended to exist
on their own but should have a host object.
"""
from __future__ import division

import honeybee.properties as hb_properties


class _Properties(object):
    """Base class for all Properties classes.

    Args:
        host: A dragonfly-core geometry object that hosts these properties
            (ie. Building, Story, Room2D).

    Properties:
        * host

    """
    _exclude = ('host', 'move', 'rotate_xy', 'reflect', 'scale', 'add_prefix',
                'reset_to_default', 'to_dict', 'to_honeybee', 'ToString')

    def __init__(self, host):
        """Initialize properties."""
        self._host = host

    @property
    def host(self):
        """Get the object hosting these properties."""
        return self._host

    @property
    def _extension_attributes(self):
        return (atr for atr in dir(self) if not atr.startswith('_')
                and atr not in self._exclude)

    def move(self, moving_vec):
        """Apply a move transform to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be moved alongside the host object.

        Args:
            moving_vec: A ladybug_geometry Vector3D with the direction and distance
                to move the face.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'move'):
                continue
            try:
                var.move(moving_vec)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to move {}: {}'.format(var, e))

    def rotate_xy(self, angle, origin):
        """Apply a rotatation in the XY plane to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be rotated alongside the host object.

        Args:
            angle: An angle in degrees.
            origin: A ladybug_geometry Point3D for the origin around which the
                object will be rotated.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'rotate_xy'):
                continue
            try:
                var.rotate_xy(angle, origin)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to rotate {}: {}'.format(var, e))

    def reflect(self, plane):
        """Apply a reflection transform to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be reflected alongside the host object.

        Args:
            plane: A ladybug_geometry Plane across which the object will
                be reflected.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'reflect'):
                continue
            try:
                var.reflect(plane)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to reflect {}: {}'.format(var, e))

    def scale(self, factor, origin=None):
        """Apply a scale transform to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be scaled alongside the host object.

        Args:
            factor: A number representing how much the object should be scaled.
            origin: A ladybug_geometry Point3D representing the origin from which
                to scale. If None, it will be scaled from the World origin (0, 0, 0).
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'scale'):
                continue
            try:
                var.scale(factor, origin)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to scale {}: {}'.format(var, e))

    def _duplicate_extension_attr(self, original_properties):
        """Duplicate the attributes added by extensions.

        This method should be called within the duplicate or __copy__ methods of
        each dragonfly-core geometry object after the core object has been duplicated.
        This method only needs to be called on the new (duplicated) core object and
        the extension properties of the original core object should be passed to
        this method as the original_properties.

        Args:
            original_properties: The properties object of the original core
                object from which the duplicate was derived.
        """
        for atr in self._extension_attributes:
            var = getattr(original_properties, atr)
            if not hasattr(var, 'duplicate'):
                continue
            try:
                setattr(self, '_' + atr, var.duplicate(self.host))
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to duplicate {}: {}'.format(var, e))

    def _add_prefix_extension_attr(self, prefix):
        """Change the identifier of attributes unique to this object by adding a prefix.

        This is particularly useful in workflows where you duplicate and edit
        a starting object and then want to combine it with the original object
        into one Model (like making a model of repeated buildings).

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the object and does not add the prefix to attributes that are
        shared across several objects.

        Args:
            prefix: Text that will be inserted at the start of the extension attributes'
                identifier. It is recommended that this prefix be short to avoid maxing
                out the 100 allowable characters for honeybee identifiers.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'add_prefix'):
                continue
            try:
                var.add_prefix(prefix)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to add prefix to {}: {}'.format(var, e))

    def _add_extension_attr_to_honeybee(self, host, honeybee_properties):
        """Add Dragonfly properties for extensions to Honeybee extension properties.

        This method should be called within the to_honeybee method for any
        dragonfly-core geometry object that maps directly to a honeybee-core object.

        Args:
            host: A honeybee-core object that hosts these properties.
            honeybee_properties: A honeybee-core Properties object to which the
                dragonfly-core extension attributes will be added.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'to_honeybee'):
                continue
            try:
                setattr(honeybee_properties, '_' + atr, var.to_honeybee(host))
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to translate {} to_honeybee: {}'.format(var, e))
        return honeybee_properties

    def _add_extension_attr_to_dict(self, base, abridged, include):
        """Add attributes for extensions to the base dictionary.

        This method should be called within the to_dict method of each dragonfly-core
        geometry object.

        Args:
            base: The dictionary of the core object without any extension
                attributes. This method will add extension attributes to this
                dictionary. For example, energy properties will appear under
                base['properties']['energy'].
            abridged: Boolean to note whether the attributes of the extensions should
                be abridged (True) or full (False). For example, if a Room's energy
                properties are abridged, the program_type attribute under the energy
                properties dictionary will just be the identifier of the program_type. If
                it is full (not abridged), the program_type will be a complete
                dictionary following the ProgramType schema. Abridged dictionaries
                should be used within the Model.to_dict but full dictionaries should
                be used within the to_dict methods of individual objects.
            include: List of properties to filter keys that must be included in
                output dictionary. For example ['energy'] will include 'energy' key if
                available in properties to_dict. By default all the keys will be
                included. To exclude all the keys from extensions use an empty list.
        """
        attr = include if include is not None else self._extension_attributes
        for atr in attr:
            var = getattr(self, atr)
            if not hasattr(var, 'to_dict'):
                continue
            try:
                base.update(var.to_dict(abridged))
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to convert {} to a dict: {}'.format(var, e))
        return base

    def _load_extension_attr_from_dict(self, property_dict):
        """Get attributes for extensions from a dictionary of the properties.

        This method should be called within the from_dict method of each dragonfly-core
        geometry object. Specifically, this method should be called on the core
        object after it has been created from a dictionary but lacks any of the
        extension attributes in the dictionary.

        Args:
            property_dict: A dictionary of properties for the object (ie.
                StoryProperties, BuildingProperties). These will be used to load
                attributes from the dictionary and assign them to the object on which
                this method is called.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'from_dict'):
                continue
            try:
                setattr(self, '_' + atr, var.__class__.from_dict(
                    property_dict[atr], self.host))
            except KeyError:
                pass  # the property_dict possesses no properties for that extension

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        """Properties representation."""
        return 'BaseProperties'


class ModelProperties(_Properties):
    """Dragonfly Model Properties. This class will be extended by extensions.

    Usage:

    .. code-block:: python

        model = Model('South Boston District', list_of_buildings)
        model.properties -> ModelProperties
        model.properties.radiance -> ModelRadianceProperties
        model.properties.energy -> ModelEnergyProperties
    """

    def to_dict(self, include=None):
        """Convert properties to dictionary.

        Args:
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'ModelProperties'}
        attr = include if include is not None else self._extension_attributes
        for atr in attr:
            var = getattr(self, atr)
            if not hasattr(var, 'to_dict'):
                continue
            try:
                base.update(var.to_dict())  # no abridged dictionary for model
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to convert {} to a dict: {}'.format(var, e))
        return base

    def apply_properties_from_dict(self, data):
        """Apply extension properties from a Model dictionary to the host Model.

        Args:
            data: A dictionary representation of an entire dragonfly-core Model.
        """
        for atr in self._extension_attributes:
            if atr not in data['properties'] or data['properties'][atr] is None:
                continue
            var = getattr(self, atr)
            if not hasattr(var, 'apply_properties_from_dict'):
                continue
            try:
                var.apply_properties_from_dict(data)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception(
                    'Failed to apply {} properties to the Model: {}'.format(atr, e))

    def to_honeybee(self, host):
        """Convert this Model's extension properties to honeybee Model properties.

        Args:
            host: A honeybee-core Model object that hosts these properties.
        """
        hb_prop = hb_properties.ModelProperties(host)
        return self._add_extension_attr_to_honeybee(host, hb_prop)

    def _check_extension_attr(self):
        """Check the attributes of extensions.

        This method should be called within the check_all method of the Model object
        to ensure that the check_all functions of any extension model properties
        are also called.
        """
        msgs = []
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'check_all'):
                continue
            try:
                msgs.append(var.check_all(raise_exception=False))
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to check_all for {}: {}'.format(var, e))
        return msgs

    def __repr__(self):
        """Properties representation."""
        return 'ModelProperties'


class ContextShadeProperties(_Properties):
    """Dragonfly ContextShade properties. This class will be extended by extensions.

    Usage:

    .. code-block:: python

        canopy = ContextShade('Outdoor Canopy', canopy_geo)
        canopy.properties -> ContextShadeProperties
        canopy.properties.radiance -> ContextShadeRadianceProperties
        canopy.properties.energy -> ContextShadeEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'ContextShadeProperties'} if not abridged else \
            {'type': 'ContextShadePropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def to_honeybee(self, host, is_mesh=False):
        """Convert this ContextShade's extension properties to honeybee Shade properties.

        Args:
            host: A honeybee-core Shade object that hosts these properties.
            is_mesh: Boolean to note whether the input host is a ShadeMesh as opposed
                to a regular Shade. (Default: False).
        """
        hb_prop = hb_properties.ShadeMeshProperties(host) \
            if is_mesh else hb_properties.ShadeProperties(host)
        return self._add_extension_attr_to_honeybee(host, hb_prop)

    def add_prefix(self, prefix):
        """Change the identifier of attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the ContextShade and does not add the prefix to attributes that are
        shared across several ContextShades.

        Args:
            prefix: Text that will be inserted at the start of extension
                attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def from_honeybee(self, hb_properties):
        """Transfer extension attributes from a Honeybee Shade to Dragonfly ContextShade.

        This method should be called within the from_honeybee method.

        Args:
            hb_properties: The properties of the honeybee Shade that is being
                translated to a Dragonfly ContextShade.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'from_honeybee') or not \
                    hasattr(hb_properties, atr):
                continue
            try:
                hb_var = getattr(hb_properties, atr)
                var.from_honeybee(hb_var)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception(
                    'Failed to translate {} from_honeybee: {}'.format(var, e))

    def __repr__(self):
        """Properties representation."""
        return 'ContextShadeProperties'


class BuildingProperties(_Properties):
    """Dragonfly Building properties. This class will be extended by extensions.

    Usage:

    .. code-block:: python

        building = Building('Office Tower', unique_stories)
        building.properties -> BuildingProperties
        building.properties.radiance -> BuildingRadianceProperties
        building.properties.energy -> BuildingEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'BuildingProperties'} if not abridged else \
            {'type': 'BuildingPropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def apply_properties_from_geojson_dict(self, data):
        """Apply extension properties to a host Building from a geoJSON dictionary.

        Args:
            data: A dictionary representation of a geoJSON feature properties.
                Specifically, this should be the "properties" key describing
                a Polygon or MultiPolygon object.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'apply_properties_from_geojson_dict'):
                continue
            try:
                var.apply_properties_from_geojson_dict(data)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception(
                    'Failed to apply {} properties to the Building: {}'.format(atr, e))

    def add_prefix(self, prefix):
        """Change the identifier of attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Building and does not add the prefix to attributes that are
        shared across several Buildings.

        Args:
            prefix: Text that will be inserted at the start of extension
                attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def make_plenums(self, room_ids):
        """Set extension attributes for Room2Ds on this host Building that are plenums.

        Args:
            room_ids: A list of identifiers for Room2Ds on this Building to be
                converted into plenums.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'make_plenums'):
                continue
            try:
                var.make_plenums(room_ids)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception(
                    'Failed to set plenum {} properties for the '
                    'Building: {}'.format(atr, e))

    def apply_ceiling_plenum_face_properties(self, room_faces, plenum_faces):
        """Set extension attributes for Honeybee Faces of ceiling plenums.

        Args:
            room_faces: A list of Honeybee Faces for the occupied Rooms that
                interface with ceiling plenums.
            plenum_faces: A list of Honeybee Faces for the ceiling plenum Rooms
                that interface with the occupied Rooms.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'apply_ceiling_plenum_face_properties'):
                continue
            try:
                var.apply_ceiling_plenum_face_properties(room_faces, plenum_faces)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception(
                    'Failed to set ceiling plenum  face {} properties for the '
                    'Building: {}'.format(atr, e))

    def apply_floor_plenum_face_properties(self, room_faces, plenum_faces):
        """Set extension attributes for Honeybee Faces of floor plenums.

        Args:
            room_faces: A list of Honeybee Faces for the occupied Rooms that
                interface with floor plenums.
            plenum_faces: A list of Honeybee Faces for the floor plenum Rooms
                that interface with the occupied Rooms.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'apply_floor_plenum_face_properties'):
                continue
            try:
                var.apply_floor_plenum_face_properties(room_faces, plenum_faces)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception(
                    'Failed to set floor plenum  face {} properties for the '
                    'Building: {}'.format(atr, e))

    def __repr__(self):
        """Properties representation."""
        return 'BuildingProperties'


class StoryProperties(_Properties):
    """Dragonfly Story properties. This class will be extended by extensions.

    Usage:

    .. code-block:: python

        story = Story('Ground Floor Retail', room_2ds)
        story.properties -> StoryProperties
        story.properties.radiance -> StoryRadianceProperties
        story.properties.energy -> StoryEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'StoryProperties'} if not abridged else \
            {'type': 'StoryPropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def add_prefix(self, prefix):
        """Change the identifier of attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Story and does not add the prefix to attributes that are
        shared across several Stories.

        Args:
            prefix: Text that will be inserted at the start of extension
                attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def __repr__(self):
        """Properties representation."""
        return 'StoryProperties'


class Room2DProperties(_Properties):
    """Dragonfly Room2D properties. This class will be extended by extensions.

    Usage:

    .. code-block:: python

        room = Room2D('Office', geometry)
        room.properties -> Room2DProperties
        room.properties.radiance -> Room2DRadianceProperties
        room.properties.energy -> Room2DEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'Room2DProperties'} if not abridged else \
            {'type': 'Room2DPropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def to_honeybee(self, host):
        """Convert this Room2D's extension properties to honeybee Room properties.

        Args:
            host: A honeybee-core Room object that hosts these properties.
        """
        hb_prop = hb_properties.RoomProperties(host)
        return self._add_extension_attr_to_honeybee(host, hb_prop)

    def add_prefix(self, prefix):
        """Change the identifier of attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Room2D (eg. single-room HVAC systems) and does not add
        the prefix to attributes that are shared across several Rooms2Ds (eg.
        ConstructionSets).

        Args:
            prefix: Text that will be inserted at the start of extension
                attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def from_honeybee(self, hb_properties):
        """Transfer extension attributes from a Honeybee Room to Dragonfly Room2D.

        This method should be called within the from_honeybee method.

        Args:
            hb_properties: The properties of the honeybee Room that is being
                translated to a Dragonfly Room2D.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'from_honeybee') or not \
                    hasattr(hb_properties, atr):
                continue
            try:
                hb_var = getattr(hb_properties, atr)
                var.from_honeybee(hb_var)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception(
                    'Failed to translate {} from_honeybee: {}'.format(var, e))

    def __repr__(self):
        """Properties representation."""
        return 'Room2DProperties'



# coding=utf-8
import pytest

from dragonfly.projection import meters_to_long_lat_factors, polygon_to_lon_lat, \
    origin_long_lat_from_location, lon_lat_to_polygon

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug.location import Location


def test_meters_to_long_lat_factors():
    """Test the meters_to_long_lat_factors method."""
    assert meters_to_long_lat_factors((0, 0))[0] == pytest.approx(111319.9, rel=1e-5)
    assert meters_to_long_lat_factors((0, 0))[1] == pytest.approx(111319.9, rel=1e-5)

    assert meters_to_long_lat_factors((0, 45))[0] < \
        meters_to_long_lat_factors((0, 45))[1] < 111319.9

    assert meters_to_long_lat_factors((0, 89))[0] < \
        meters_to_long_lat_factors((0, 89))[1] < 111319.9


def test_polygon_to_lon_lat():
    """Test the polygon_to_lon_lat method."""
    polygon = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    lat = 42.0
    lon = -70.0
    verts1 = polygon_to_lon_lat(polygon, (lon, lat))

    assert len(verts1) == 4
    for vert in verts1:
        assert vert[0] == pytest.approx(lon, rel=1e-5)
        assert vert[1] == pytest.approx(lat, rel=1e-5)

    convert_facs = meters_to_long_lat_factors((lon, lat))
    verts2 = polygon_to_lon_lat(polygon, (lon, lat), convert_facs)

    for vert1, verts2 in zip(verts1, verts2):
        assert vert1 == verts2


def test_origin_long_lat_from_location():
    """Test the origin_long_lat_from_location method."""
    lat = 42.0
    lon = -70.0
    loc = Location(latitude=lat, longitude=lon)
    o_lon, o_lat = origin_long_lat_from_location(loc, Point2D(10, 10))

    assert lat == pytest.approx(o_lat, rel=1e-5)
    assert lon == pytest.approx(o_lon, rel=1e-5)


def test_lon_lat_to_polygon():
    """Test conversion of lon lat to model units"""

    polygon_lon_lat_coords = [
        (-70.0, 42.0),
        (-69.99997578750273, 42.0),
        (-69.99997578750273, 42.00001799339205),
        (-70.0, 42.00001799339205)]

    polygon = lon_lat_to_polygon(
        polygon_lon_lat_coords, origin_lon_lat=(-70.0, 42.0))

    test_polygon = [Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2)]

    # Check length
    assert len(test_polygon) == len(polygon)

    # Check coordinate values
    for point, test_point in zip(polygon, test_polygon):
        assert test_point[0] == pytest.approx(point[0], abs=1e-5)
        assert test_point[1] == pytest.approx(point[1], abs=1e-5)



"""dragonfly model creation commands."""
import click
import sys
import logging
import json

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug.location import Location
from honeybee.units import UNITS_TOLERANCES
from honeybee.model import Model as HBModel
from dragonfly.windowparameter import SimpleWindowRatio
from dragonfly.model import Model

_logger = logging.getLogger(__name__)


@click.group(help='Commands for creating Dragonfly models.')
def create():
    pass


@create.command('merge-models')
@click.argument('base-model', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option(
    '--other-model', '-m', help='The other Model to be merged into the base model.',
    type=click.File('w'), multiple=True)
@click.option(
    '--output-file', '-f', help='Optional file to output the Model DFJSON string. '
    'By default it will be printed out to stdout',
    type=click.File('w'), default='-')
def merge_models_cli(base_model, other_model, output_file):
    """Create a Dragonfly Model by merging multiple models together.

    \b
    Args:
        base_model: Full path to a Dragonfly Model JSON or Pkl file that serves
            as the base into which the other model(s) will be merged. This model
            determines the units and tolerance of the output model.
    """
    try:
        merge_models(base_model, other_model, output_file)
    except Exception as e:
        _logger.exception('Model merging failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


def merge_models(base_model, other_model=(), output_file=None):
    """Create a Dragonfly Model by merging multiple models together.

    Args:
        base_model: Full path to a Dragonfly Model JSON or Pkl file that serves
            as the base into which the other model(s) will be merged. This model
            determines the units and tolerance of the output model.
        other_model: A list of other Dragonfly Models to be merged into the base model.
        output_file: Optional file to output the Model DFJSON string. If None,
            the string will be returned from this method. (Default: None).
    """
    # serialize the Model and convert the units
    parsed_model = Model.from_file(base_model)
    other_models = [Model.from_file(m) for m in other_model]
    for o_model in other_models:
        parsed_model.add_model(o_model)

    # write the new model out to the file or stdout
    if output_file is None:
        return json.dumps(parsed_model.to_dict())
    output_file.write(json.dumps(parsed_model.to_dict()))


@create.command('from-honeybee')
@click.argument('base-model', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option(
    '--conversion-method', '-c', help='Text to indicate how the Honeybee Rooms '
    'should be converted to Dragonfly. Choose from: AllRoom2D, ExtrudedOnly, AllRoom3D. '
    'Note that the AllRoom2D option may result in some loss or simplification of '
    'the 3D Honeybee geometry but ensures that all of the Dragonfly features for '
    'editing the rooms can be used. The ExtrudedOnly method will convert only the '
    '3D Rooms that would have no loss or simplification of geometry when converted '
    'to Room2D. AllRoom3D keeps all detailed 3D geometry on the Building.room_3ds '
    'property, enabling you to convert the 3D Rooms to Room2D using the '
    'Building.convert_room_3ds_to_2d() method as you see fit.',
    type=str, default="ExtrudedOnly", show_default=True)
@click.option(
    '--other-model', '-m', help='Another Honeybee Model to be added as a separate '
    'Building in the resulting Dragonfly Model.',
    type=click.File('w'), multiple=True)
@click.option(
    '--output-file', '-f', help='Optional file to output the Model DFJSON string. '
    'By default it will be printed out to stdout',
    type=click.File('w'), default='-')
def from_honeybee_cli(base_model, conversion_method, other_model, output_file):
    """Create a Dragonfly Model from Honeybee Model(s).

    \b
    Args:
        base_model: Full path to a Honeybee Model JSON or Pkl file that serves
            as the base for the resulting Dragonfly Model. This model
            determines the units and tolerance of the output model.
    """
    try:
        from_honeybee(base_model, conversion_method, other_model, output_file)
    except Exception as e:
        _logger.exception('Model creation from honeybee failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


def from_honeybee(base_model, conversion_method='ExtrudedOnly',
                  other_model=(), output_file=None):
    """Create a Dragonfly Model from Honeybee Model(s).

    Args:
        base_model: Full path to a Honeybee Model JSON or Pkl file that serves
            as the base for the resulting Dragonfly Model. This model
            determines the units and tolerance of the output model.
        conversion_method: Text to indicate how the Honeybee Rooms should be
            converted to Dragonfly. Choose from: AllRoom2D, ExtrudedOnly, AllRoom3D.
            Note that the AllRoom2D option may result in some loss or simplification
            of the 3D Honeybee geometry but ensures that all of the Dragonfly
            features for editing the rooms can be used. The ExtrudedOnly method
            will convert only the 3D Rooms that would have no loss or
            simplification of geometry when converted to Room2D. AllRoom3D
            keeps all detailed 3D geometry on the Building.room_3ds property,
            enabling you to convert the 3D Rooms to Room2D using the
            Building.convert_room_3ds_to_2d() method as you see fit.
        other_model: An optional list of other Honeybee Models to be added as
            a separate Building in the resulting Dragonfly Model.
        output_file: Optional file to output the Model DFJSON string. If None,
            the string will be returned from this method. (Default: None).
    """
    # serialize the input Model(s)
    hb_model = HBModel.from_file(base_model)
    other_models = [HBModel.from_file(m) for m in other_model]

    # convert the Honeybee Model(s) to Dragonfly
    df_model = Model.from_honeybee(hb_model, conversion_method)
    for o_hb_model in other_models:
        o_df_model = Model.from_honeybee(o_hb_model, conversion_method)
        df_model.add_model(o_df_model)

    # write the new model out to the file or stdout
    if output_file is None:
        return json.dumps(df_model.to_dict())
    output_file.write(json.dumps(df_model.to_dict()))


@create.command('from-geojson')
@click.argument('geojson', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--location', '-l', help='An optional latitude and longitude, formatted '
              'as two floats separated by a comma, (eg. "42.3601,-71.0589"), defining '
              'the origin of the geojson file. If nothing is passed, the origin is '
              'autocalculated as the bottom-left corner of the bounding box of all '
              'building footprints in the geojson file.', type=str, default=None)
@click.option('--point', '-p', help='An optional X and Y coordinate, formatted '
              'as two floats separated by a comma, (eg. "200,200"), defining '
              'the origin of the geojson file in the space of the dragonfly model. The '
              'coordinates of this point are expected to be in the expected units '
              'of this Model.', type=str, default='0,0', show_default=True)
@click.option('--window-ratio', '-wr', help='A number between 0 and 1 (but not equal '
              'to 1) for the ratio between aperture area and wall area to be applied to '
              'all walls of all buildings. If specified, this will override the '
              'window_ratio key in the geojson.', type=float, default=None)
@click.option('--buildings-only/--all-to-buildings', ' /-all', help='Flag to indicate '
              'if all geometries in the geojson file should be considered buildings. '
              'If buildings-only, this method will only generate footprints from '
              'geometries that are defined as a "Building" in the type field of its '
              'corresponding properties.', default=True, show_default=True)
@click.option('--no-context/--existing-to-context', ' /-c', help='Flag to indicate '
              'whether polygons possessing a building_status of "Existing" under their '
              'properties should be imported as ContextShade instead of Building '
              'objects.', default=True, show_default=True)
@click.option('--separate-top-bottom/--no-separation', ' /-ns', help='Flag to indicate '
              'whether top/bottom stories of the buildings should not be separated in '
              'order to account for different boundary conditions of the roof and '
              'ground floor.', default=True, show_default=True)
@click.option('--units', '-u', help='Text for the units system in which the '
              'resulting model geometry should be. Must be (Meters, Millimeters, '
              'Feet, Inches, Centimeters).',
              type=str, default='Meters', show_default=True)
@click.option('--tolerance', '-t', help='The maximum difference between x, y, and z '
              'values at which vertices are considered equivalent.',
              type=float, default=None)
@click.option('--output-file', '-f', help='Optional file to output the Model JSON '
              'string. By default it will be printed out to stdout',
              type=click.File('w'), default='-')
def from_geojson_cli(
        geojson, location, point, window_ratio, buildings_only,
        no_context, separate_top_bottom, units, tolerance, output_file):
    """Create a Dragonfly model from a geojson file with building footprints.

    \b
    Args:
        geojson: Full path to a geoJSON file with building footprints as Polygons
            or MultiPolygons.
    """
    try:
        all_to_buildings = not buildings_only
        existing_to_context = not no_context
        no_separation = not separate_top_bottom
        from_geojson(
            geojson, location, point, window_ratio, all_to_buildings,
            existing_to_context, no_separation, units, tolerance, output_file)
    except Exception as e:
        _logger.exception('Model creation from geoJSON failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


def from_geojson(
        geojson, location=None, point=None, window_ratio=None,
        all_to_buildings=False, existing_to_context=False, no_separation=False,
        units='Meters', tolerance=None, output_file=None,
        buildings_only=True, no_context=True, separate_top_bottom=True):
    """Create a Dragonfly model from a geojson file with building footprints.

    Args:
        geojson: Full path to a geoJSON file with building footprints as Polygons
            or MultiPolygons.
        location: An optional latitude and longitude, formatted as two floats
            separated by a comma, (eg. "42.3601,-71.0589"), defining the origin
            of the geojson file. If None, the origin is autocalculated as the
            bottom-left corner of the bounding box of all building footprints
            in the geojson file.
        point: An optional X and Y coordinate, formatted as two floats separated
            by a comma, (eg. "200,200"), defining the origin of the geojson file
            in the space of the dragonfly model. The coordinates of this point
            are expected to be in the expected units of this Model.
        window_ratio: A number between 0 and 1 (but not equal to 1) for the
            ratio between aperture area and wall area to be applied to all walls
            of all buildings. If specified, this will override the window_ratio
            key in the geojson.
        all_to_buildings: Boolean to indicate if all geometries in the geojson
            file should be considered buildings. If buildings-only, this method
            will only generate footprints from geometries that are defined as
            a "Building" in the type field of its corresponding properties.
        existing_to_context: Boolean to indicate whether polygons possessing a
            building_status of "Existing" under their properties should be
            imported as ContextShade instead of Building objects.
        no_separation: Boolean to indicate whether polygons possessing a
            building_status of "Existing" under their properties should be
            imported as ContextShade instead of Building objects.
        units: Text for the units system in which the resulting model geometry
            should be. Must be (Meters, Millimeters, Feet, Inches, Centimeters).
        tolerance: The maximum difference between x, y, and z values at which
            vertices are considered equivalent.
        output_file: Optional file to output the Model JSON string. If None,
            The string will be returned from this method.
    """
    # parse the location and point if they are specified
    if location is not None:
        lat, lon = [float(num) for num in location.split(',')]
        location = Location(longitude=lat, latitude=lon)
    if point is None:
        point = Point2D(0, 0)
    else:
        point = Point2D(*[float(num) for num in point.split(',')])

    # create the model object
    tolerance = tolerance if tolerance is not None else UNITS_TOLERANCES[units]
    model, _ = Model.from_geojson(
        geojson, location, point, all_to_buildings, existing_to_context,
        units=units, tolerance=tolerance)

    # apply windows if the window ratio and top/bottom separation if specified
    if window_ratio is not None:
        model.set_outdoor_window_parameters(SimpleWindowRatio(window_ratio))
    if not no_separation:
        model.separate_top_bottom_floors()

    # write the model out to the file or stdout
    if output_file is None:
        return json.dumps(model.to_dict())
    output_file.write(json.dumps(model.to_dict()))



# coding: utf-8
"""Skylight Parameters with instructions for generating skylights."""
from __future__ import division
import math
import sys
if (sys.version_info < (3, 0)):  # python 2
    from itertools import izip as zip  # python 2

from ladybug_geometry.geometry2d import Point2D, Vector2D, Polygon2D
from ladybug_geometry.geometry3d import Vector3D, Point3D, Face3D
from ladybug_geometry.bounding import bounding_rectangle

from honeybee.typing import float_in_range, float_positive
from honeybee.altnumber import autocalculate
from honeybee.aperture import Aperture
from honeybee.door import Door


class _SkylightParameterBase(object):
    """Base object for all Skylight parameters.

    This object records the minimum number of the methods that must be overwritten
    on a skylight parameter object for it to be successfully be applied in
    dragonfly workflows.
    """
    __slots__ = ()

    def __init__(self):
        pass

    def area_from_face(self, face):
        """Get the skylight area generated by these parameters from a Room2D Face3D."""
        return 0

    def add_skylight_to_face(self, face, tolerance=0.01):
        """Add Apertures to a Honeybee Roof Face using these Skylight Parameters."""
        pass

    def scale(self, factor):
        """Get a scaled version of these SkylightParameters.

        This method is called within the scale methods of the Room2D.

        Args:
            factor: A number representing how much the object should be scaled.
        """
        return self

    def split(self, face_3ds, tolerance=0.01):
        """Split these skylight parameters across several Face3D.

        This is useful when the Room2D to which the skylight is assigned has been
        split into multiple Room2Ds.

        Args:
            face_3ds: A list of horizontal Face3D representing the floor geometries
                of a Room2D to which these skylight parameters are split across.
            tolerance: The maximum difference between point values for them to be
                considered distinct. (Default: 0.01, suitable for objects in meters).
        """
        return self  # assume even distribution by default

    @classmethod
    def from_dict(cls, data):
        """Create SkylightParameterBase from a dictionary.

        .. code-block:: python

            {
            "type": "SkylightParameterBase"
            }
        """
        assert data['type'] == 'SkylightParameterBase', \
            'Expected SkylightParameterBase dictionary. Got {}.'.format(data['type'])
        return cls()

    def to_dict(self):
        """Get SkylightParameterBase as a dictionary."""
        return {'type': 'SkylightParameterBase'}

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def ToString(self):
        return self.__repr__()

    def __copy__(self):
        return _SkylightParameterBase()

    def __repr__(self):
        return 'SkylightParameterBase'


class GriddedSkylightArea(_SkylightParameterBase):
    """Instructions for gridded skylights defined by an absolute area.

    Args:
        skylight_area: A number for the skylight area in current model units.
            If this area is larger than the area of the roof that it is applied
            to, the skylight will fill the parent roof at a 99% ratio.
        spacing: A number for the spacing between the centers of each grid cell.
            This should be less than a third of the dimension of the Roof geometry
            if multiple, evenly-spaced skylights are desired. If None or Autocalculate,
            a spacing of one third the smaller dimension of the parent Roof will
            be automatically assumed. (Default: Autocalculate).

    Properties:
        * skylight_area
        * spacing
    """
    __slots__ = ('_skylight_area', '_spacing')

    def __init__(self, skylight_area, spacing=autocalculate):
        """Initialize GriddedSkylightArea."""
        self._skylight_area = float_positive(skylight_area, 'skylight area')
        if spacing == autocalculate:
            spacing = None
        elif spacing is not None:
            spacing = float_positive(spacing, 'skylight spacing')
            assert spacing > 0, 'GriddedSkylightArea spacing must be greater than zero.'
        self._spacing = spacing

    @property
    def skylight_area(self):
        """Get a number for the skylight area in current model units."""
        return self._skylight_area

    @property
    def spacing(self):
        """Get a number or the spacing between the skylights.

        None indicates that the spacing will always be one third of the smaller
        dimension of the parent Roof.
        """
        return self._spacing

    def area_from_face(self, face):
        """Get the skylight area generated by these parameters from a Room2D Face3D.

        Args:
            face: A Roof Face3D to which these parameters are applied.
        """
        return self._skylight_area

    def add_skylight_to_face(self, face, tolerance=0.01):
        """Add Apertures to a Honeybee Roof Face using these Skylight Parameters.

        Args:
            face: A honeybee-core Face object.
            tolerance: The maximum difference between point values for them to be
                considered distinct. (Default: 0.01, suitable for objects in meters).
        """
        if self._skylight_area == 0:
            return None
        spacing = self.spacing
        if self.spacing is None:
            min_pt, max_pt = face.min, face.max
            min_dim = min(max_pt.x - min_pt.x, max_pt.y - min_pt.y)
            spacing = (min_dim / 3) - tolerance
        skylight_ratio = self.skylight_area / face.area
        skylight_ratio = 0.99 if skylight_ratio > 0.99 else skylight_ratio
        face.apertures_by_ratio_gridded(
            skylight_ratio, spacing, tolerance=tolerance)

    def scale(self, factor):
        """Get a scaled version of these SkylightParameters.

        This method is called within the scale methods of the Room2D.

        Args:
            factor: A number representing how much the object should be scaled.
        """
        spc = self.spacing * factor if self.spacing is not None else None
        return GriddedSkylightArea(self.skylight_area * factor, spc)

    def split(self, face_3ds, tolerance=0.01):
        """Split these skylight parameters across several Face3D.

        This is useful when the Room2D to which the skylight is assigned has been
        split into multiple Room2Ds.

        Args:
            face_3ds: A list of horizontal Face3D representing the floor geometries
                of a Room2D to which these skylight parameters are split across.
            tolerance: The maximum difference between point values for them to be
                considered distinct. (Default: 0.01, suitable for objects in meters).
        """
        total_area = sum(face.area for face in face_3ds)
        return [GriddedSkylightArea(self.area * (face.area / total_area), self.spacing)
                for face in face_3ds]

    @classmethod
    def from_dict(cls, data):
        """Create GriddedSkylightArea from a dictionary.

        .. code-block:: python

            {
            "type": "GriddedSkylightArea",
            "skylight_area": 2.5,
            "spacing": 2
            }
        """
        assert data['type'] == 'GriddedSkylightArea', \
            'Expected GriddedSkylightArea dictionary. Got {}.'.format(data['type'])
        spc = data['spacing'] if 'spacing' in data and \
            data['spacing'] != autocalculate.to_dict() else None
        return cls(data['skylight_area'], spc)

    def to_dict(self):
        """Get GriddedSkylightArea as a dictionary."""
        base = {'type': 'GriddedSkylightArea', 'skylight_area': self.skylight_area}
        if self.spacing is not None:
            base['spacing'] = self.spacing
        return base

    def __copy__(self):
        return GriddedSkylightArea(
            self.skylight_area, self.spacing)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.skylight_area, self.spacing)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, GriddedSkylightArea) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'GriddedSkylightArea: [area: {}]'.format(self.skylight_area)


class GriddedSkylightRatio(_SkylightParameterBase):
    """Instructions for gridded skylights derived from an area ratio with the roof.

    Args:
        skylight_ratio: A number between 0 and 1 for the ratio between the skylight
            area and the total Roof face area.
        spacing: A number for the spacing between the centers of each grid cell.
            This should be less than a third of the dimension of the Roof geometry
            if multiple, evenly-spaced skylights are desired. If None or Autocalculate,
            a spacing of one third the smaller dimension of the parent Roof will
            be automatically assumed. (Default: Autocalculate).

    Properties:
        * skylight_ratio
        * spacing
    """
    __slots__ = ('_skylight_ratio', '_spacing')

    def __init__(self, skylight_ratio, spacing=autocalculate):
        """Initialize GriddedSkylightRatio."""
        self._skylight_ratio = float_in_range(skylight_ratio, 0, 1.0, 'skylight ratio')
        if spacing == autocalculate:
            spacing = None
        elif spacing is not None:
            spacing = float_positive(spacing, 'skylight spacing')
            assert spacing > 0, 'GriddedSkylightRatio spacing must be greater than zero.'
        self._spacing = spacing

    @property
    def skylight_ratio(self):
        """Get a number between 0 and 1 for the skylight ratio."""
        return self._skylight_ratio

    @property
    def spacing(self):
        """Get a number or the spacing between the skylights.

        None indicates that the spacing will always be one third of the smaller
        dimension of the parent Roof.
        """
        return self._spacing

    def area_from_face(self, face):
        """Get the skylight area generated by these parameters from a Room2D Face3D.

        Args:
            face: A Roof Face3D to which these parameters are applied.
        """
        return face.area * self._skylight_ratio

    def add_skylight_to_face(self, face, tolerance=0.01):
        """Add Apertures to a Honeybee Roof Face using these Skylight Parameters.

        Args:
            face: A honeybee-core Face object.
            tolerance: The maximum difference between point values for them to be
                considered distinct. (Default: 0.01, suitable for objects in meters).
        """
        if self._skylight_ratio == 0:
            return None
        spacing = self.spacing
        if self.spacing is None:
            min_pt, max_pt = face.min, face.max
            min_dim = min(max_pt.x - min_pt.x, max_pt.y - min_pt.y)
            spacing = (min_dim / 3) - tolerance
        face.apertures_by_ratio_gridded(
            self.skylight_ratio, spacing, tolerance=tolerance)

    def scale(self, factor):
        """Get a scaled version of these SkylightParameters.

        This method is called within the scale methods of the Room2D.

        Args:
            factor: A number representing how much the object should be scaled.
        """
        spc = self.spacing * factor if self.spacing is not None else None
        return GriddedSkylightRatio(self.skylight_ratio, spc)

    @classmethod
    def from_dict(cls, data):
        """Create GriddedSkylightRatio from a dictionary.

        .. code-block:: python

            {
            "type": "GriddedSkylightRatio",
            "skylight_ratio": 0.05,
            "spacing": 2
            }
        """
        assert data['type'] == 'GriddedSkylightRatio', \
            'Expected GriddedSkylightRatio dictionary. Got {}.'.format(data['type'])
        spc = data['spacing'] if 'spacing' in data and \
            data['spacing'] != autocalculate.to_dict() else None
        return cls(data['skylight_ratio'], spc)

    def to_dict(self):
        """Get GriddedSkylightRatio as a dictionary."""
        base = {'type': 'GriddedSkylightRatio', 'skylight_ratio': self.skylight_ratio}
        if self.spacing is not None:
            base['spacing'] = self.spacing
        return base

    def __copy__(self):
        return GriddedSkylightRatio(
            self.skylight_ratio, self.spacing)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.skylight_ratio, self.spacing)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, GriddedSkylightRatio) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'GriddedSkylightRatio: [ratio: {}]'.format(self.skylight_ratio)


class DetailedSkylights(_SkylightParameterBase):
    """Instructions for detailed skylights, defined by 2D Polygons (lists of 2D vertices).

    Note that these parameters are intended to represent skylights that are specific
    to a particular Room2D and, unlike the other SkylightsParameters, this class
    performs no automatic checks to ensure that the skylights lie within the
    boundary of the Roof that they are assigned to.

    Args:
        polygons: An array of ladybug_geometry Polygon2D objects in world XY
            coordinates with one polygon for each skylight. These coordinate
            values should lie within the Room2D's Polygon2D.
        are_doors: An array of booleans that align with the polygons and note whether
            each of the polygons represents an overhead door (True) or a skylight
            (False). If None, it will be assumed that all polygons represent skylights
            and they will be translated to Apertures in any resulting Honeybee
            model. (Default: None).

    Properties:
        * polygons
        * are_doors
    """
    __slots__ = ('_polygons', '_are_doors')

    def __init__(self, polygons, are_doors=None):
        """Initialize DetailedSkylights."""
        if not isinstance(polygons, tuple):
            polygons = tuple(polygons)
        for polygon in polygons:
            assert isinstance(polygon, Polygon2D), \
                'Expected Polygon2D for skylight polygon. Got {}'.format(type(polygon))
        assert len(polygons) != 0, \
            'There must be at least one polygon to use DetailedSkylights.'
        self._polygons = polygons
        if are_doors is None:
            self._are_doors = (False,) * len(polygons)
        else:
            if not isinstance(are_doors, tuple):
                are_doors = tuple(are_doors)
            for is_dr in are_doors:
                assert isinstance(is_dr, bool), 'Expected booleans for ' \
                    'DetailedSkylights.are_doors. Got {}'.format(type(is_dr))
            assert len(are_doors) == len(polygons), \
                'Length of DetailedSkylights.are_doors ({}) does not match the length ' \
                'of DetailedSkylights.polygons ({}).'.format(
                    len(are_doors), len(polygons))
            self._are_doors = are_doors

    @property
    def polygons(self):
        """Get an array of Polygon2Ds with one polygon for each skylight."""
        return self._polygons

    @property
    def are_doors(self):
        """Get an array of booleans that note whether each polygon is a door."""
        return self._are_doors

    def area_from_face(self, face):
        """Get the sub-face area generated by these parameters from a Room2D Face3D.

        This includes both skylights and overhead doors.

        Args:
            face: A Roof Face3D to which these parameters are applied.
        """
        return sum(polygon.area for polygon in self._polygons)

    def aperture_area_from_face(self, face):
        """Get the Aperture area generated by these parameters from a Room2D Face3D.

        Args:
            face: A Roof Face3D to which these parameters are applied.
        """
        return sum(poly.area for poly, isd in zip(self.polygons, self.are_doors)
                   if not isd)

    def check_overlaps(self, tolerance=0.01):
        """Check whether any polygons overlap with one another.

        Args:
            tolerance: The minimum distance that two polygons must overlap in order
                for them to be considered overlapping and invalid. (Default: 0.01,
                suitable for objects in meters).

        Returns:
            A string with the message. Will be an empty string if valid.
        """
        # group the polygons according to their overlaps
        grouped_polys = Polygon2D.group_by_overlap(self.polygons, tolerance)
        # report any polygons that overlap
        if not all(len(g) == 1 for g in grouped_polys):
            base_msg = '({} skylights overlap one another)'
            all_msg = []
            for p_group in grouped_polys:
                if len(p_group) != 1:
                    all_msg.append(base_msg.format(len(p_group)))
            return ' '.join(all_msg)
        return ''

    def check_self_intersecting(self, tolerance=0.01):
        """Check whether any polygons in these skylight parameters are self intersecting.

        Args:
            tolerance: The minimum distance between a vertex coordinates where
                they are considered equivalent. (Default: 0.01, suitable
                for objects in meters).

        Returns:
            A string with the message. Will be an empty string if valid.
        """
        self_int_i = []
        for i, polygon in enumerate(self.polygons):
            if polygon.is_self_intersecting:
                try:
                    new_geo = polygon.remove_colinear_vertices(tolerance)
                    if new_geo.is_self_intersecting:
                        self_int_i.append(str(i))
                except AssertionError:
                    self_int_i.append(str(i))
        if len(self_int_i) != 0:
            return 'Skylight polygons with the following indices are ' \
                'self-intersecting: ({})'.format(' '.join(self_int_i))
        return ''

    def check_valid_for_face(self, face):
        """Check that these skylight parameters are valid for a given Face3D.

        Args:
            face: A Roof Face3D to which these parameters are applied.

        Returns:
            A string with the message. Will be an empty string if valid.
        """
        # first check that the total skylight area isn't larger than the roof
        total_area = face.area
        win_area = self.area_from_face(face)
        if win_area >= total_area:
            return 'Total area of skylights [{}] is greater than the area of the ' \
                'parent roof [{}].'.format(win_area, total_area)
        # next, check to be sure that no skylight is out of the roof boundary
        msg_template = 'Skylight polygon {} is outside the range allowed ' \
            'by the parent roof.'
        verts2d = tuple(Point2D(pt.x, pt.y) for pt in face.boundary)
        parent_poly, parent_holes = Polygon2D(verts2d), None
        if face.has_holes:
            parent_holes = tuple(
                Polygon2D(Point2D(pt.x, pt.y) for pt in hole)
                for hole in face.holes
            )
        for i, p_gon in enumerate(self.polygons):
            if not self._is_sub_polygon(p_gon, parent_poly, parent_holes):
                return msg_template.format(i)
        return ''

    def add_skylight_to_face(self, face, tolerance=0.01):
        """Add Apertures to a Honeybee Roof Face using these Skylight Parameters.

        Args:
            face: A honeybee-core Face object.
            tolerance: The maximum difference between point values for them to be
                considered distinct. (Default: 0.01, suitable for objects in meters).
        """
        # get the polygons that represent the roof face
        verts2d = tuple(Point2D(pt.x, pt.y) for pt in face.geometry.boundary)
        parent_poly, parent_holes = Polygon2D(verts2d), None
        if face.geometry.has_holes:
            parent_holes = tuple(
                Polygon2D(Point2D(pt.x, pt.y) for pt in hole)
                for hole in face.geometry.holes
            )
        # loop through each polygon and create its geometry
        sub_count = 1
        p_dir = Vector3D(0, 0, 1)
        for polygon, isd in zip(self.polygons, self.are_doors):
            poly_relation = self._sub_poly_relation(
                polygon, parent_poly, parent_holes, tolerance)
            if poly_relation == -1:
                continue
            pt3d = tuple(
                face.geometry.plane.project_point(Point3D(p.x, p.y, 0), p_dir)
                for p in polygon)
            if None in pt3d:
                continue
            s_geos = [Face3D(pt3d)]
            if poly_relation == 0:  # find the boolean difference
                par_geo = face.geometry
                bool_int = Face3D.coplanar_intersection(
                    par_geo, s_geos[0], tolerance, math.radians(1))
                if bool_int is None or len(bool_int) == 0:  # boolean intersection failed
                    continue
                # offset the result of the boolean intersection from the edge
                s_geos = []
                for res in bool_int:
                    new_pts = [res.plane.xy_to_xyz(pt2) for pt2 in
                               res.boundary_polygon2d.offset(tolerance * 2)]
                    s_geos.append(Face3D(new_pts))
            for s_geo in s_geos:
                if isd:
                    sub_f = Door('{}_Door{}'.format(face.identifier, sub_count), s_geo)
                    face.add_door(sub_f)
                else:
                    sub_f = Aperture('{}_Glz{}'.format(face.identifier, sub_count), s_geo)
                    face.add_aperture(sub_f)
                sub_count += 1

    def move(self, moving_vec):
        """Get this DetailedSkylights moved along a vector.

        Args:
            moving_vec: A Vector3D with the direction and distance to move the polygon.
        """
        vec2 = Vector2D(moving_vec.x, moving_vec.y)
        return DetailedSkylights(
            tuple(polygon.move(vec2) for polygon in self.polygons),
            self.are_doors)

    def scale(self, factor, origin=None):
        """Get a scaled version of these DetailedSkylights.

        This method is called within the scale methods of the Room2D.

        Args:
            factor: A number representing how much the object should be scaled.
            origin: A ladybug_geometry Point3D representing the origin from which
                to scale. If None, it will be scaled from the World origin (0, 0, 0).
        """
        ori = Point2D(origin.x, origin.y) if origin is not None else None
        return DetailedSkylights(
            tuple(polygon.scale(factor, ori) for polygon in self.polygons),
            self.are_doors)

    def rotate(self, angle, origin):
        """Get these DetailedSkylights rotated counterclockwise in the XY plane.

        Args:
            angle: An angle in degrees.
            origin: A ladybug_geometry Point3D for the origin around which the
                object will be rotated.
        """
        ori, ang = Point2D(origin.x, origin.y), math.radians(angle)
        return DetailedSkylights(
            tuple(polygon.rotate(ang, ori) for polygon in self.polygons),
            self.are_doors)

    def reflect(self, plane):
        """Get a reflected version of these DetailedSkylights across a plane.

        Args:
            plane: A ladybug_geometry Plane across which the object will be reflected.
        """
        # get the plane normal and origin in 2D space
        normal = Vector2D(plane.n.x, plane.n.y)
        origin = Point2D(plane.o.x, plane.o.y)
        # loop through the polygons and reflect them across the plane
        new_polygons = []
        for polygon in self.polygons:
            new_verts = tuple(pt.reflect(normal, origin) for pt in polygon.vertices)
            new_polygons.append(Polygon2D(tuple(reversed(new_verts))))
        return DetailedSkylights(new_polygons, self.are_doors)

    def split(self, face_3ds, tolerance=0.01):
        """Split these skylight parameters across several Face3D.

        This is useful when the Room2D to which the skylight is assigned has been
        split into multiple Room2Ds.

        Args:
            face_3ds: A list of horizontal Face3D representing the floor geometries
                of a Room2D to which these skylight parameters are split across.
            tolerance: The maximum difference between point values for them to be
                considered distinct. (Default: 0.01, suitable for objects in meters).
        """
        # loop through the polygons and build up new skylight parameters
        new_sky_par = []
        for face_3d in face_3ds:
            verts2d = tuple(Point2D(pt.x, pt.y) for pt in face_3d.boundary)
            room_poly = Polygon2D(verts2d)
            rel_poly, rel_doors, offset_needed = [], [], False
            for sky_poly, isd in zip(self.polygons, self.are_doors):
                poly_rel = room_poly.polygon_relationship(sky_poly, tolerance)
                if poly_rel >= 0:
                    rel_poly.append(sky_poly)
                    rel_doors.append(isd)
                    if poly_rel == 0:
                        offset_needed = True
            if len(rel_poly) == 0:
                new_sky_par.append(None)
                continue
            new_sky = DetailedSkylights(rel_poly, rel_doors)
            if offset_needed:
                new_sky.offset_polygons_for_face(face_3d, tolerance, tolerance)
            new_sky_par.append(new_sky)
        return new_sky_par

    def offset(self, offset_distance, tolerance=0.01):
        """Offset the edges of all skylight polygons by a certain distance.

        This is useful for translating between interfaces that expect the window
        frame to be included within or excluded from the geometry.

        Note that this operation can often create polygons that collide with
        one another or extend past the parent Face. So it may be desirable
        to run the union_overlaps method after using this one.

        Args:
            offset_distance: Distance with which the edges of each polygon will
                be offset from the original geometry. Positive values will
                offset the geometry outwards and negative values will offset the
                geometries inwards.
            tolerance: The minimum difference between point values for them to be
                considered the distinct. (Default: 0.01, suitable for objects
                in meters).
        """
        offset_polys, offset_are_doors = [], []
        for polygon, isd in zip(self.polygons, self.are_doors):
            try:
                poly = polygon.remove_colinear_vertices(tolerance)
                off_poly = poly.offset(-offset_distance, False)
                if not off_poly.is_self_intersecting:
                    offset_polys.append(off_poly)
                else:  # polygon became self-intersecting after offset
                    offset_polys.append(poly)
                offset_are_doors.append(isd)
            except AssertionError:  # degenerate window to ignore
                pass
        self._polygons = tuple(offset_polys)
        self._are_doors = tuple(offset_are_doors)

    def offset_polygons_for_face(self, face_3d, offset_distance=0.05, tolerance=0.01):
        """Offset the polygons until all vertices are inside the boundaries of a Face3D.

        Args:
            face_3d: A horizontal Face3D representing the floor geometry of a Room2D
                to which these skylight parameters are assigned.
            offset_distance: Distance from the edge of the face_3d that
                the polygons will be offset to. (Default: 0.05, suitable for
                objects in meters).
            tolerance: The maximum difference between point values for them to be
                considered distinct. (Default: 0.01, suitable for objects in meters).
        """
        # get the polygons that represent the roof face
        face_z = face_3d[0].z
        verts2d = tuple(Point2D(pt.x, pt.y) for pt in face_3d.boundary)
        parent_poly, parent_holes = Polygon2D(verts2d), None
        if face_3d.has_holes:
            parent_holes = tuple(
                Polygon2D(Point2D(pt.x, pt.y) for pt in hole)
                for hole in face_3d.holes
            )
        # loop through the polygons and offset them if they are not correctly bounded
        new_polygons, new_are_doors = [], []
        for polygon, isd in zip(self.polygons, self.are_doors):
            if not self._is_sub_polygon(polygon, parent_poly, parent_holes):
                # find the boolean intersection of the polygon with the room
                sub_face = Face3D([Point3D.from_point2d(pt, face_z) for pt in polygon])
                bool_int = Face3D.coplanar_intersection(
                    face_3d, sub_face, tolerance, math.radians(1))
                if bool_int is None:  # skylight completely outside parent
                    continue
                # offset the result of the boolean intersection from the edge
                parent_edges = face_3d.boundary_segments if face_3d.holes is None \
                    else face_3d.boundary_segments + \
                    tuple(seg for hole in face_3d.hole_segments for seg in hole)
                for new_f in bool_int:
                    new_pts_2d = []
                    for pt in new_f.boundary:
                        for edge in parent_edges:
                            close_pt = edge.closest_point(pt)
                            if pt.distance_to_point(close_pt) < offset_distance:
                                move_vec = edge.v.rotate_xy(math.pi / 2).normalize()
                                move_vec = move_vec * offset_distance
                                pt = pt.move(move_vec)
                        new_pts_2d.append(Point2D(pt.x, pt.y))
                    new_polygons.append(Polygon2D(new_pts_2d))
                    new_are_doors.append(isd)
            else:
                new_polygons.append(polygon)
                new_are_doors.append(isd)
        # assign the offset polygons to this object
        self._polygons = tuple(new_polygons)
        self._are_doors = tuple(new_are_doors)

    def remove_self_intersecting(self, tolerance=0.01):
        """Remove any skylight polygons that are self intersecting.

        Args:
            tolerance: The minimum distance between a vertex coordinates where
                they are considered equivalent. (Default: 0.01, suitable
                for objects in meters).

        Returns:
            A string with the message. Will be an empty string if valid.
        """
        new_polygons, new_are_doors = [], []
        for polygon, isd in zip(self.polygons, self.are_doors):
            if not polygon.is_self_intersecting:
                new_polygons.append(polygon)
                new_are_doors.append(isd)
            else:
                try:
                    new_geo = polygon.remove_colinear_vertices(tolerance)
                    if not new_geo.is_self_intersecting:
                        new_polygons.append(new_geo)
                        new_are_doors.append(isd)
                except AssertionError:
                    pass
        # assign the clean polygons to this object
        self._polygons = tuple(new_polygons)
        self._are_doors = tuple(new_are_doors)

    def union_overlaps(self, tolerance=0.01):
        """Union any skylight polygons that overlap with one another.

        Args:
            tolerance: The minimum distance that two polygons must overlap in order
                for them to be considered overlapping. (Default: 0.01,
                suitable for objects in meters).
        """
        # group the polygons by their overlap
        grouped_polys = Polygon2D.group_by_overlap(self.polygons, tolerance)
        # union any of the polygons that overlap
        if not all(len(g) == 1 for g in grouped_polys):
            new_polys = []
            for p_group in grouped_polys:
                if len(p_group) == 1:
                    new_polys.append(p_group[0])
                else:
                    union_poly = Polygon2D.boolean_union_all(p_group, tolerance)
                    for new_poly in union_poly:
                        new_polys.append(new_poly.remove_colinear_vertices(tolerance))
            self._reassign_are_doors(new_polys)
            self._polygons = tuple(new_polys)

    def merge_and_simplify(self, max_separation, tolerance=0.01):
        """Merge skylight polygons that are close to one another into a single polygon.

        This can be used to create a simpler set of skylights that is easier to
        edit and is in the same location as the original skylights.

        Args:
            max_separation: A number for the maximum distance between skylight polygons
                at which point they will be merged into a single geometry. Typically,
                this max_separation should be set to a value that is slightly larger
                than the window frame. Setting this equal to the tolerance will
                simply join neighboring skylights together.
            tolerance: The maximum difference between point values for them to be
                considered distinct. (Default: 0.01, suitable for objects in meters).
        """
        # gather a clean version of the polygons with colinear vertices removed
        clean_polys = []
        for poly in self.polygons:
            try:
                clean_polys.append(poly.remove_colinear_vertices(tolerance))
            except AssertionError:  # degenerate geometry to ignore
                pass
        # join the polygons together
        if max_separation <= tolerance:
            new_polys = Polygon2D.joined_intersected_boundary(
                clean_polys, tolerance)
        else:
            new_polys = Polygon2D.gap_crossing_boundary(
                clean_polys, max_separation, tolerance)
        self._reassign_are_doors(new_polys)
        self._polygons = tuple(new_polys)

    def merge_to_bounding_rectangle(self, tolerance=0.01):
        """Merge skylight polygons that touch or overlap with one another to a rectangle.

        Args:
            tolerance: The minimum distance from the edge of a neighboring polygon
                at which a point is considered to touch that polygon. (Default: 0.01,
                suitable for objects in meters).
        """
        # group the polygons by their overlap
        grouped_polys = Polygon2D.group_by_touching(self.polygons, tolerance)
        # union any of the polygons that overlap
        if not all(len(g) == 1 for g in grouped_polys):
            new_polys = []
            for p_group in grouped_polys:
                if len(p_group) == 1:
                    new_polys.append(p_group[0])
                else:
                    min_pt, max_pt = bounding_rectangle(p_group)
                    rect_verts = (
                        min_pt, Point2D(max_pt.x, min_pt.y),
                        max_pt, Point2D(min_pt.x, max_pt.y))
                    rect_poly = Polygon2D(rect_verts)
                    new_polys.append(rect_poly)
            self._reassign_are_doors(new_polys)
            self._polygons = tuple(new_polys)

    def remove_small_skylights(self, area_threshold):
        """Remove any small skylight polygons that are below a certain area threshold.

        Args:
            area_threshold: A number for the area below which a skylight polygon
                will be removed.
        """
        new_polygons, new_are_doors = [], []
        for poly, is_dr in zip(self.polygons, self.are_doors):
            if poly.area > area_threshold:
                new_polygons.append(poly)
                new_are_doors.append(is_dr)
        self._polygons = tuple(new_polygons)
        self._are_doors = tuple(new_are_doors)

    @classmethod
    def from_dict(cls, data):
        """Create DetailedSkylights from a dictionary.

        Args:
            data: A dictionary in the format below.

        .. code-block:: python

            {
            "type": "DetailedSkylights",
            "polygons": [((0.5, 0.5), (2, 0.5), (2, 2), (0.5, 2)),
                         ((3, 1), (4, 1), (4, 2))],
            "are_doors": [False]
            }
        """
        assert data['type'] == 'DetailedSkylights', \
            'Expected DetailedSkylights dictionary. Got {}.'.format(data['type'])
        are_doors = data['are_doors'] if 'are_doors' in data else None
        return cls(
            tuple(Polygon2D(tuple(Point2D.from_array(pt) for pt in poly))
                  for poly in data['polygons']),
            are_doors
        )

    def to_dict(self):
        """Get DetailedSkylights as a dictionary."""
        return {
            'type': 'DetailedSkylights',
            'polygons': [[pt.to_array() for pt in poly] for poly in self.polygons],
            'are_doors': self.are_doors
        }

    def _reassign_are_doors(self, new_polys):
        """Reset the are_doors property using a set of new polygons."""
        if len(new_polys) != len(self._polygons):
            if all(not dr for dr in self._are_doors):  # common case of no overhead doors
                self._are_doors = (False,) * len(new_polys)
            else:
                new_are_doors = []
                for n_poly in new_polys:
                    for o_poly, is_door in zip(self.polygons, self.are_doors):
                        if n_poly.is_point_inside_bound_rect(o_poly.center):
                            new_are_doors.append(is_door)
                            break
                    else:
                        new_are_doors.append(False)
                self._are_doors = tuple(new_are_doors)

    @staticmethod
    def _is_sub_polygon(sub_poly, parent_poly, parent_holes=None):
        """Check if a sub-polygon is valid for a given assumed parent polygon.

        Args:
            sub_poly: A sub-Polygon2D for which sub-face equivalency will be tested.
            parent_poly: A parent Polygon2D.
            parent_holes: An optional list of Polygon2D for any holes that may
                exist in the parent polygon. (Default: None).
        """
        if parent_holes is None:
            return parent_poly.is_polygon_inside(sub_poly)
        else:
            if not parent_poly.is_polygon_inside(sub_poly):
                return False
            for hole_poly in parent_holes:
                if not hole_poly.is_polygon_outside(sub_poly):
                    return False
            return True

    @staticmethod
    def _sub_poly_relation(sub_polygon, parent_poly, parent_holes, tolerance):
        """Check the relationship between a sub_polygon and a parent polygon.

        Args:
            sub_polygon: A Polygon2D which will be checked if it lies inside the parent.
            parent_poly: A parent Polygon2D.
            parent_holes: An optional list of Polygon2D for any holes that may
                exist in the parent polygon. (Default: None).
        """
        base_relation = parent_poly.polygon_relationship(sub_polygon, tolerance)
        if parent_holes is None:
            return base_relation
        else:
            if base_relation == -1:
                return -1
            for hole_poly in parent_holes:
                hole_relation = hole_poly.polygon_relationship(sub_polygon, tolerance)
                if hole_relation == 1:
                    return -1
            return base_relation

    @staticmethod
    def _is_sub_point(sub_point, parent_poly, parent_holes=None):
        """Check if a point lies inside a parent polygon.

        Args:
            sub_point: A Point2D which will be checked if it lies inside the parent.
            parent_poly: A parent Polygon2D.
            parent_holes: An optional list of Polygon2D for any holes that may
                exist in the parent polygon. (Default: None).
        """
        if parent_holes is None:
            return parent_poly.is_point_inside(sub_point)
        else:
            if not parent_poly.is_point_inside(sub_point):
                return False
            for hole_poly in parent_holes:
                if hole_poly.is_point_inside(sub_point):
                    return False
            return True

    def __len__(self):
        return len(self._polygons)

    def __getitem__(self, key):
        return self._polygons[key]

    def __iter__(self):
        return iter(self._polygons)

    def __copy__(self):
        return DetailedSkylights(self._polygons, self._are_doors)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return tuple(hash(polygon) for polygon in self._polygons) + self.are_doors

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, DetailedSkylights) and \
            len(self._polygons) == len(other._polygons)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'DetailedSkylights: [{} windows]'.format(len(self._polygons))
